/********************************************************************
*  File: CPUGausSeidel.c
*
*  Description:
*	
*	Mainfunction to compute an inverse Matrix from a positive definite 
*   Matrix. The Routine is using the Gaus Seidel Matrix invertion algorithm.
*   
*   1   2   1   |  1  0   0               1   0   0  |  -2.5   1.5   0.5
*   2   3   1   |  0  1   0       =>      0   1   0  |   1.5  -0.5  -0.5 
*   1   1   2   |  0  0   1               0   0   1  |   0.5  -0.5   0.5 
*   Inputmatrix       E                       E          Inverse Matrix
*
*  Arguments: 
*	- float *hDataIn      Input Matrix 1D, no data changes
*   - float *hDataOut     Output Matrix 1D, the inverse datamatrix  
*   - int size            Matrix dimension in size, width = height = size
*	
* Used custom Routines:
*	- padDiag ()
*   - calcL ()
*   - calcR ()
*   - normalizeDiag ()
*
*********************************************************************/

#include <malloc.h>
#include <memory.h>

#include "../Header/CPUGausSeidel.h"


int CPUGausSeidel (float *hDataIn, float *hDataOut, int size)
{
    int i;
    float *paddedMatrix;
    float *tmpPaddedMatrix;
    int stride = 2 * size;

    //Memory allocation for the Inputdata and identity matrix in one array.
    paddedMatrix = (float *) malloc (2 * size * size * sizeof (float));

    if (paddedMatrix == 0)
        return -1;

    //Store the Inputdata with the identity matrix
    padDiag (hDataIn, paddedMatrix, size, stride);

    //get the lower left diagonal matrix to zero, 1st step of Gaus elimination
    calcL (paddedMatrix, size, stride);
    //get the rigth upper diagonal matrix to zero, 2nd step of Gaus elimination
    calcR (paddedMatrix, size, stride);

    //Normalize the diagonal entries to get the identity matrix on the left side
    normalizeDiag (paddedMatrix, size, stride);

    //write the inverse Matrix back
    tmpPaddedMatrix = paddedMatrix;
    for (i = 0; i < size; i++)
    {
        memcpy (hDataOut, tmpPaddedMatrix + size, size * sizeof (float));
        hDataOut += size;
        tmpPaddedMatrix += stride;
    }

    free (paddedMatrix);

    return 0;
}
